/*******************************************************************************
 * This file is part of the Zivid API
 *
 * Copyright 2015-2025 (C) Zivid AS
 * All rights reserved.
 *
 * Zivid Software License, v1.0
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of Zivid AS nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 4. This software, with or without modification, must not be used with any
 * other 3D camera than from Zivid AS.
 *
 * 5. Any software provided in binary form under this license must not be
 * reverse engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY ZIVID AS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL ZIVID AS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contact: Zivid Customer Success Team <customersuccess@zivid.com>
 * Info:    http://www.zivid.com
 ******************************************************************************/

#pragma once

#include "Zivid/Detail/UtilsExport.h"

#include <memory>
#include <string>

namespace Zivid
{

#ifndef NO_DOC
    /// This namespace is for implementation details, and should not be relied on by users of the Zivid API.
    namespace Detail
    {
        ZIVID_UTILS_EXPORT std::string array1DToString(std::size_t size);
    } // namespace Detail
#endif

    /// <summary>One-dimensional container of data</summary>
    /// <remarks>Array1D manages the data using implicit sharing via reference counting. Copy
    /// or assignment of an Array1D object is shallow (only a pointer to the data is copied),
    /// which makes copy/assignment of Array1D fast. The data stored in the array is deallocated
    /// when there are no Array1D instances pointing to it.
    ///
    /// The type of the elements stored in the array is given by template parameter DataFormat.
    /// </remarks>
    template<typename DataFormat>
    class Array1D
    {
    public:
        /// <summary>The type of the elements stored in the Array</summary>
        using ValueType = DataFormat;

        /// <summary>
        /// The iterator type for immutable access. It iterates over individual Array1D elements.
        /// </summary>
        using ConstIterator = const DataFormat *;

        /// <summary>Create an empty Array1D</summary>
        Array1D()
            : m_size{ 0 }
        {}

        /// <summary>Get the number of elements in the array</summary>
        size_t size() const
        {
            return m_size;
        }

        /// <summary>Check if the array is empty</summary>
        bool isEmpty() const
        {
            return size() == 0;
        }

        /// <summary>Pointer to the first data element of the array</summary>
        /// <remarks>If the array is empty then this method returns nullptr</remarks>
        const DataFormat *data() const
        {
            return m_data.get();
        }

        /// <summary>Constant reference to an element given by an index</summary>
        /// <remarks>
        /// No bounds checking is performed. Calling this  operator with an argument that is out of range
        /// causes undefined behavior. Calling this operator when the Array1D is empty causes undefined behavior.
        /// </remarks>
        /// <param name="idx">Index (from 0 to <see cref="size" /> - 1)</param>
        const DataFormat &operator()(size_t idx) const
        {
#ifdef __clang__
#    if __has_warning("-Wunsafe-buffer-usage")
#        pragma clang diagnostic push
#        pragma clang diagnostic ignored "-Wunsafe-buffer-usage"
#    endif
#endif
            return *(data() + idx);
#ifdef __clang__
#    if __has_warning("-Wunsafe-buffer-usage")
#        pragma clang diagnostic pop
#    endif
#endif
        }

        /// <summary>Iterator to the beginning of the array</summary>
        ConstIterator begin() const
        {
            return data();
        }

        /// <summary>Iterator to the end of the array</summary>
        ConstIterator end() const
        {
#ifdef __clang__
#    if __has_warning("-Wunsafe-buffer-usage")
#        pragma clang diagnostic push
#        pragma clang diagnostic ignored "-Wunsafe-buffer-usage"
#    endif
#endif
            return data() + size();
#ifdef __clang__
#    if __has_warning("-Wunsafe-buffer-usage")
#        pragma clang diagnostic pop
#    endif
#endif
        }

        /// <summary>Iterator to the beginning of the array</summary>
        ConstIterator cbegin() const
        {
            return begin();
        }

        /// <summary>Iterator to the end of the array</summary>
        ConstIterator cend() const
        {
            return end();
        }

        /// <summary>Get array information as string</summary>
        std::string toString() const
        {
            return Detail::array1DToString(size());
        }

    private:
        friend class Array1DFactory;

        Array1D(std::size_t size, std::unique_ptr<DataFormat[]> data)
            : m_size{ size }
            , m_data{ std::move(data) }
        {}

        size_t m_size;
        std::shared_ptr<const DataFormat[]> m_data;
    };

    /// <summary>Serialize array information to a stream</summary>
    template<typename T>
    std::ostream &operator<<(std::ostream &stream, const Array1D<T> &array)
    {
        return stream << array.toString();
    }

    /// <summary>Iterator to the beginning of the array</summary>
    template<typename T>
    typename Array1D<T>::ConstIterator cbegin(const Array1D<T> &array)
    {
        return array.cbegin();
    }

    /// <summary>Iterator to the beginning of the array</summary>
    template<typename T>
    typename Array1D<T>::ConstIterator begin(const Array1D<T> &array)
    {
        return array.begin();
    }

    /// <summary>Iterator to the end of the array</summary>
    template<typename T>
    typename Array1D<T>::ConstIterator cend(const Array1D<T> &array)
    {
        return array.cend();
    }

    /// <summary>Iterator to the end of the array</summary>
    template<typename T>
    typename Array1D<T>::ConstIterator end(const Array1D<T> &array)
    {
        return array.end();
    }
} // namespace Zivid
