/*******************************************************************************
 * This file is part of the Zivid API
 *
 * Copyright 2015-2025 (C) Zivid AS
 * All rights reserved.
 *
 * Zivid Software License, v1.0
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of Zivid AS nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 4. This software, with or without modification, must not be used with any
 * other 3D camera than from Zivid AS.
 *
 * 5. Any software provided in binary form under this license must not be
 * reverse engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY ZIVID AS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL ZIVID AS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contact: Zivid Customer Success Team <customersuccess@zivid.com>
 * Info:    http://www.zivid.com
 ******************************************************************************/

#pragma once

#include <array>
#include <chrono>
#include <cmath>
#include <ctime>
#include <iomanip>
#include <memory>
#include <set>
#include <sstream>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

#include "Zivid/DataModel/Detail/Detail.h"
#include "Zivid/DataModel/NodeType.h"
#include "Zivid/DataModel/Traits.h"
#include "Zivid/Detail/CoreExport.h"
#include "Zivid/Detail/TypeTraits.h"
#include "Zivid/Range.h"

#ifdef _MSC_VER
#    pragma warning(push)
#    pragma warning(disable : 4251) // "X needs to have dll-interface to be used by clients of class Y."
#endif

namespace Zivid
{
    namespace Experimental
    {

        /// <summary>Input parameters for controlling the local point cloud registration process.
        /// </summary>

        // NOLINTNEXTLINE(clang-analyzer-optin.performance.Padding)
        class ZIVID_CORE_EXPORT LocalPointCloudRegistrationParameters
        {
        public:
            /// <summary>The type of this node</summary>
            static constexpr DataModel::NodeType nodeType = DataModel::NodeType::group;

            /// <summary>The full path for this value</summary>
            static constexpr const char *path{ "" };

            /// <summary>The name of this value</summary>
            static constexpr const char *name{ "LocalPointCloudRegistrationParameters" };

            /// <summary>The description for this value</summary>
            static constexpr const char *description{
                R"description(Input parameters for controlling the local point cloud registration process.
)description"
            };

            static constexpr size_t version{ 1 };

#ifndef NO_DOC
            template<size_t>
            struct Version;

            using LatestVersion = Zivid::Experimental::LocalPointCloudRegistrationParameters;

            // Short identifier. This value is not guaranteed to be universally unique
            // Todo(ZIVID-2808): Move this to internal DataModelExt header
            static constexpr std::array<uint8_t, 3> binaryId{ 'l', 'p', 'r' };

#endif

            /// <summary>Criteria for ending iteration early. Iteration will end when all criteria are satisfied.
            /// </summary>

            // NOLINTNEXTLINE(clang-analyzer-optin.performance.Padding)
            class ZIVID_CORE_EXPORT ConvergenceCriteria
            {
            public:
                /// <summary>The type of this node</summary>
                static constexpr DataModel::NodeType nodeType = DataModel::NodeType::group;

                /// <summary>The full path for this value</summary>
                static constexpr const char *path{ "ConvergenceCriteria" };

                /// <summary>The name of this value</summary>
                static constexpr const char *name{ "ConvergenceCriteria" };

                /// <summary>The description for this value</summary>
                static constexpr const char *description{
                    R"description(Criteria for ending iteration early. Iteration will end when all criteria are satisfied.
)description"
                };

                /// <summary>Stop iteration when the absolute change in root mean square error (RMSE) from previous iteration is less than this threshold.
                /// RMSE is defined as the square root of the average squared distance between corresponding points in the source and target point clouds.
                /// </summary>

                // NOLINTNEXTLINE(clang-analyzer-optin.performance.Padding)
                class ZIVID_CORE_EXPORT RMSEDiffThreshold
                {
                public:
                    /// <summary>The type of this node</summary>
                    static constexpr DataModel::NodeType nodeType = DataModel::NodeType::leafValue;

                    /// <summary>The full path for this value</summary>
                    static constexpr const char *path{ "ConvergenceCriteria/RMSEDiffThreshold" };

                    /// <summary>The name of this value</summary>
                    static constexpr const char *name{ "RMSEDiffThreshold" };

                    /// <summary>The description for this value</summary>
                    static constexpr const char *description{
                        R"description(Stop iteration when the absolute change in root mean square error (RMSE) from previous iteration is less than this threshold.
RMSE is defined as the square root of the average squared distance between corresponding points in the source and target point clouds.
)description"
                    };

                    /// <summary>The type of the underlying value</summary>
                    using ValueType = float;

                    /// <summary>The range of valid values for RMSEDiffThreshold</summary>
                    static constexpr Range<float> validRange()
                    {
                        return { std::numeric_limits<float>::lowest(), std::numeric_limits<float>::max() };
                    }

                    /// <summary>Default constructor</summary>
                    RMSEDiffThreshold() = default;

                    /// <summary>Constructor</summary>
                    explicit constexpr RMSEDiffThreshold(float value)
                        : m_value{ value }
                    {}

                    /// <summary>Get the value</summary>
                    float value() const;

                    /// <summary>Get the value as string</summary>
                    std::string toString() const;

                    /// <summary>Comparison operator</summary>
                    bool operator==(const RMSEDiffThreshold &other) const
                    {
                        return m_value == other.m_value;
                    }

                    /// <summary>Comparison operator</summary>
                    bool operator!=(const RMSEDiffThreshold &other) const
                    {
                        return m_value != other.m_value;
                    }

                    /// <summary>Comparison operator</summary>
                    bool operator<(const RMSEDiffThreshold &other) const
                    {
                        return m_value < other.m_value;
                    }

                    /// <summary>Comparison operator</summary>
                    bool operator>(const RMSEDiffThreshold &other) const
                    {
                        return m_value > other.m_value;
                    }

                    /// <summary>Comparison operator</summary>
                    bool operator<=(const RMSEDiffThreshold &other) const
                    {
                        return m_value <= other.m_value;
                    }

                    /// <summary>Comparison operator</summary>
                    bool operator>=(const RMSEDiffThreshold &other) const
                    {
                        return m_value >= other.m_value;
                    }

                    /// <summary>Operator to serialize the value to a stream</summary>
                    friend std::ostream &operator<<(std::ostream &stream, const RMSEDiffThreshold &value)
                    {
                        return stream << value.toString();
                    }

                private:
                    void setFromString(const std::string &value);

                    float m_value{ 1e-4F };

                    friend struct DataModel::Detail::Befriend<RMSEDiffThreshold>;
                };

                /// <summary>Stop iteration when the absolute change in coverage from previous iteration is less than this threshold.
                /// Coverage is defined as the fraction of points in the source point cloud that found a correspondence in the target point cloud.
                /// </summary>

                // NOLINTNEXTLINE(clang-analyzer-optin.performance.Padding)
                class ZIVID_CORE_EXPORT SourceCoverageDiffThreshold
                {
                public:
                    /// <summary>The type of this node</summary>
                    static constexpr DataModel::NodeType nodeType = DataModel::NodeType::leafValue;

                    /// <summary>The full path for this value</summary>
                    static constexpr const char *path{ "ConvergenceCriteria/SourceCoverageDiffThreshold" };

                    /// <summary>The name of this value</summary>
                    static constexpr const char *name{ "SourceCoverageDiffThreshold" };

                    /// <summary>The description for this value</summary>
                    static constexpr const char *description{
                        R"description(Stop iteration when the absolute change in coverage from previous iteration is less than this threshold.
Coverage is defined as the fraction of points in the source point cloud that found a correspondence in the target point cloud.
)description"
                    };

                    /// <summary>The type of the underlying value</summary>
                    using ValueType = float;

                    /// <summary>The range of valid values for SourceCoverageDiffThreshold</summary>
                    static constexpr Range<float> validRange()
                    {
                        return { std::numeric_limits<float>::lowest(), std::numeric_limits<float>::max() };
                    }

                    /// <summary>Default constructor</summary>
                    SourceCoverageDiffThreshold() = default;

                    /// <summary>Constructor</summary>
                    explicit constexpr SourceCoverageDiffThreshold(float value)
                        : m_value{ value }
                    {}

                    /// <summary>Get the value</summary>
                    float value() const;

                    /// <summary>Get the value as string</summary>
                    std::string toString() const;

                    /// <summary>Comparison operator</summary>
                    bool operator==(const SourceCoverageDiffThreshold &other) const
                    {
                        return m_value == other.m_value;
                    }

                    /// <summary>Comparison operator</summary>
                    bool operator!=(const SourceCoverageDiffThreshold &other) const
                    {
                        return m_value != other.m_value;
                    }

                    /// <summary>Comparison operator</summary>
                    bool operator<(const SourceCoverageDiffThreshold &other) const
                    {
                        return m_value < other.m_value;
                    }

                    /// <summary>Comparison operator</summary>
                    bool operator>(const SourceCoverageDiffThreshold &other) const
                    {
                        return m_value > other.m_value;
                    }

                    /// <summary>Comparison operator</summary>
                    bool operator<=(const SourceCoverageDiffThreshold &other) const
                    {
                        return m_value <= other.m_value;
                    }

                    /// <summary>Comparison operator</summary>
                    bool operator>=(const SourceCoverageDiffThreshold &other) const
                    {
                        return m_value >= other.m_value;
                    }

                    /// <summary>Operator to serialize the value to a stream</summary>
                    friend std::ostream &operator<<(std::ostream &stream, const SourceCoverageDiffThreshold &value)
                    {
                        return stream << value.toString();
                    }

                private:
                    void setFromString(const std::string &value);

                    float m_value{ 1e-4F };

                    friend struct DataModel::Detail::Befriend<SourceCoverageDiffThreshold>;
                };

                using Descendants = std::tuple<
                    LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold,
                    LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold>;

                /// <summary>Default constructor</summary>
                ConvergenceCriteria();

                /// <summary>Constructor taking variadic number of arguments</summary>
                /// <remarks>
                /// One or more descendant types can be provided. All types not provided will be set to their default
                /// value. The result is the same as default constructing `ConvergenceCriteria` then invoking `set(args)`.
                ///
                /// The provided arguments must be descendants of `ConvergenceCriteria`. These types can be provided:
                ///
                /// <list type="bullet">
                /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold</item>
                /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold</item>
                /// </list>
                /// </remarks>
#ifndef NO_DOC
                template<
                    typename... Args,
                    typename std::enable_if<sizeof...(Args) >= 1, int>::type = 0,
                    typename std::enable_if<
                        Zivid::Detail::TypeTraits::AllArgsAreInTuple<Descendants, typename std::decay<Args>::type...>::
                            value,
                        int>::type = 0>
#else
                template<typename... Args>
#endif
                explicit ConvergenceCriteria(Args &&...args)
                {
                    using namespace Zivid::Detail::TypeTraits;

                    static_assert(
                        AllArgsDecayedAreUnique<Args...>::value,
                        "Found duplicate types among the arguments passed to ConvergenceCriteria(...). "
                        "Types should be listed at most once.");

                    set(std::forward<Args>(args)...);
                }

                /// <summary>Set multiple arguments</summary>
                /// <remarks>
                /// The method invokes `set(arg)` with each of the provided arguments.
                ///
                /// The provided arguments must be descendants of `ConvergenceCriteria`. These types can be provided:
                ///
                /// <list type="bullet">
                /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold</item>
                /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold</item>
                /// </list>
                /// </remarks>
#ifndef NO_DOC
                template<typename... Args, typename std::enable_if<sizeof...(Args) >= 2, int>::type = 0>
#else
                template<typename... Args>
#endif
                void set(Args &&...args)
                {
                    using namespace Zivid::Detail::TypeTraits;

                    using AllArgsAreDescendantNodes =
                        AllArgsAreInTuple<Descendants, typename std::decay<Args>::type...>;
                    static_assert(
                        AllArgsAreDescendantNodes::value, "All arguments passed to set(...) must be descendant nodes.");

                    static_assert(
                        AllArgsDecayedAreUnique<Args...>::value,
                        "Found duplicate types among the arguments passed to set(...). "
                        "Types should be listed at most once.");

                    Zivid::DataModel::Detail::invokeSetWithEachArgument(*this, std::forward<Args>(args)...);
                }

                /// <summary>Returns a copy of this object with the given argument(s) set to the new value(s)</summary>
                /// <remarks>
                /// Creates a copy of this ConvergenceCriteria object, then invokes `set(args)` on the copy, and
                /// finally returns the copy. This method does not modify the original object.
                ///
                /// The provided arguments must be descendants of `ConvergenceCriteria`. These types can be provided:
                ///
                /// <list type="bullet">
                /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold</item>
                /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold</item>
                /// </list>
                /// </remarks>
#ifndef NO_DOC
                template<typename... Args, typename std::enable_if<sizeof...(Args) >= 1, int>::type = 0>
#else
                template<typename... Args>
#endif
                ConvergenceCriteria copyWith(Args &&...args) const
                {
                    using namespace Zivid::Detail::TypeTraits;

                    using AllArgsAreDescendantNodes =
                        AllArgsAreInTuple<Descendants, typename std::decay<Args>::type...>;
                    static_assert(
                        AllArgsAreDescendantNodes::value,
                        "All arguments passed to copyWith(...) must be descendant nodes.");

                    static_assert(
                        AllArgsDecayedAreUnique<Args...>::value,
                        "Found duplicate types among the arguments passed to copyWith(...). "
                        "Types should be listed at most once.");

                    auto copy{ *this };
                    copy.set(std::forward<Args>(args)...);
                    return copy;
                }

                /// <summary>Get RMSEDiffThreshold</summary>
                const RMSEDiffThreshold &rmseDiffThreshold() const
                {
                    return m_rmseDiffThreshold;
                }

                /// <summary>Get RMSEDiffThreshold</summary>
                RMSEDiffThreshold &rmseDiffThreshold()
                {
                    return m_rmseDiffThreshold;
                }

                /// <summary>Set RMSEDiffThreshold</summary>
                ConvergenceCriteria &set(const RMSEDiffThreshold &value)
                {
                    m_rmseDiffThreshold = value;
                    return *this;
                }

                /// <summary>Get SourceCoverageDiffThreshold</summary>
                const SourceCoverageDiffThreshold &sourceCoverageDiffThreshold() const
                {
                    return m_sourceCoverageDiffThreshold;
                }

                /// <summary>Get SourceCoverageDiffThreshold</summary>
                SourceCoverageDiffThreshold &sourceCoverageDiffThreshold()
                {
                    return m_sourceCoverageDiffThreshold;
                }

                /// <summary>Set SourceCoverageDiffThreshold</summary>
                ConvergenceCriteria &set(const SourceCoverageDiffThreshold &value)
                {
                    m_sourceCoverageDiffThreshold = value;
                    return *this;
                }

                template<
                    typename T,
                    typename std::enable_if<
                        std::is_same<T, LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold>::
                            value,
                        int>::type = 0>
                const LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold &get() const
                {
                    return m_rmseDiffThreshold;
                }

                template<
                    typename T,
                    typename std::enable_if<
                        std::is_same<
                            T,
                            LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold>::
                            value,
                        int>::type = 0>
                const LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold &get()
                    const
                {
                    return m_sourceCoverageDiffThreshold;
                }

                template<size_t i, typename std::enable_if<i == 0, int>::type = 0>
                const LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold &get() const
                {
                    return m_rmseDiffThreshold;
                }

                template<size_t i, typename std::enable_if<i == 1, int>::type = 0>
                const LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold &get()
                    const
                {
                    return m_sourceCoverageDiffThreshold;
                }

                /// <summary>Run the given function on each direct member with the value of the member as parameter</summary>
                template<typename F>
                void forEach(const F &f) const
                {
                    f(m_rmseDiffThreshold);
                    f(m_sourceCoverageDiffThreshold);
                }

                /// <summary>Run the given function on each direct member with the value of the member as parameter</summary>
                template<typename F>
                void forEach(const F &f)
                {
                    f(m_rmseDiffThreshold);
                    f(m_sourceCoverageDiffThreshold);
                }

                /// <summary>Equality operator</summary>
                bool operator==(const ConvergenceCriteria &other) const;

                /// <summary>Inequality operator</summary>
                bool operator!=(const ConvergenceCriteria &other) const;

                /// <summary>Get the value as string</summary>
                std::string toString() const;

                /// <summary>Operator to send the value as string to a stream</summary>
                friend std::ostream &operator<<(std::ostream &stream, const ConvergenceCriteria &value)
                {
                    return stream << value.toString();
                }

            private:
                void setFromString(const std::string &value);

                void setFromString(const std::string &fullPath, const std::string &value);

                std::string getString(const std::string &fullPath) const;

                RMSEDiffThreshold m_rmseDiffThreshold;
                SourceCoverageDiffThreshold m_sourceCoverageDiffThreshold;

                friend struct DataModel::Detail::Befriend<ConvergenceCriteria>;
            };

            /// <summary>The radius around each source point in which the algorithm will look for neighbors in the target point cloud.
            /// Should be larger than the typical distance between points in the point clouds.
            /// This parameter may have a significant performance impact.
            /// </summary>

            // NOLINTNEXTLINE(clang-analyzer-optin.performance.Padding)
            class ZIVID_CORE_EXPORT MaxCorrespondenceDistance
            {
            public:
                /// <summary>The type of this node</summary>
                static constexpr DataModel::NodeType nodeType = DataModel::NodeType::leafValue;

                /// <summary>The full path for this value</summary>
                static constexpr const char *path{ "MaxCorrespondenceDistance" };

                /// <summary>The name of this value</summary>
                static constexpr const char *name{ "MaxCorrespondenceDistance" };

                /// <summary>The description for this value</summary>
                static constexpr const char *description{
                    R"description(The radius around each source point in which the algorithm will look for neighbors in the target point cloud.
Should be larger than the typical distance between points in the point clouds.
This parameter may have a significant performance impact.
)description"
                };

                /// <summary>The type of the underlying value</summary>
                using ValueType = float;

                /// <summary>The range of valid values for MaxCorrespondenceDistance</summary>
                static constexpr Range<float> validRange()
                {
                    return { std::numeric_limits<float>::lowest(), std::numeric_limits<float>::max() };
                }

                /// <summary>Default constructor</summary>
                MaxCorrespondenceDistance() = default;

                /// <summary>Constructor</summary>
                explicit constexpr MaxCorrespondenceDistance(float value)
                    : m_value{ value }
                {}

                /// <summary>Get the value</summary>
                float value() const;

                /// <summary>Get the value as string</summary>
                std::string toString() const;

                /// <summary>Comparison operator</summary>
                bool operator==(const MaxCorrespondenceDistance &other) const
                {
                    return m_value == other.m_value;
                }

                /// <summary>Comparison operator</summary>
                bool operator!=(const MaxCorrespondenceDistance &other) const
                {
                    return m_value != other.m_value;
                }

                /// <summary>Comparison operator</summary>
                bool operator<(const MaxCorrespondenceDistance &other) const
                {
                    return m_value < other.m_value;
                }

                /// <summary>Comparison operator</summary>
                bool operator>(const MaxCorrespondenceDistance &other) const
                {
                    return m_value > other.m_value;
                }

                /// <summary>Comparison operator</summary>
                bool operator<=(const MaxCorrespondenceDistance &other) const
                {
                    return m_value <= other.m_value;
                }

                /// <summary>Comparison operator</summary>
                bool operator>=(const MaxCorrespondenceDistance &other) const
                {
                    return m_value >= other.m_value;
                }

                /// <summary>Operator to serialize the value to a stream</summary>
                friend std::ostream &operator<<(std::ostream &stream, const MaxCorrespondenceDistance &value)
                {
                    return stream << value.toString();
                }

            private:
                void setFromString(const std::string &value);

                float m_value{ 10.0F };

                friend struct DataModel::Detail::Befriend<MaxCorrespondenceDistance>;
            };

            /// <summary>End iteration when reaching this number, even if convergence criteria are not yet satisfied.
            /// </summary>

            // NOLINTNEXTLINE(clang-analyzer-optin.performance.Padding)
            class ZIVID_CORE_EXPORT MaxIterationCount
            {
            public:
                /// <summary>The type of this node</summary>
                static constexpr DataModel::NodeType nodeType = DataModel::NodeType::leafValue;

                /// <summary>The full path for this value</summary>
                static constexpr const char *path{ "MaxIterationCount" };

                /// <summary>The name of this value</summary>
                static constexpr const char *name{ "MaxIterationCount" };

                /// <summary>The description for this value</summary>
                static constexpr const char *description{
                    R"description(End iteration when reaching this number, even if convergence criteria are not yet satisfied.
)description"
                };

                /// <summary>The type of the underlying value</summary>
                using ValueType = uint32_t;

                /// <summary>The range of valid values for MaxIterationCount</summary>
                static constexpr Range<uint32_t> validRange()
                {
                    return { std::numeric_limits<uint32_t>::lowest(), std::numeric_limits<uint32_t>::max() };
                }

                /// <summary>Default constructor</summary>
                MaxIterationCount() = default;

                /// <summary>Constructor</summary>
                explicit constexpr MaxIterationCount(uint32_t value)
                    : m_value{ value }
                {}

                /// <summary>Get the value</summary>
                uint32_t value() const;

                /// <summary>Get the value as string</summary>
                std::string toString() const;

                /// <summary>Comparison operator</summary>
                bool operator==(const MaxIterationCount &other) const
                {
                    return m_value == other.m_value;
                }

                /// <summary>Comparison operator</summary>
                bool operator!=(const MaxIterationCount &other) const
                {
                    return m_value != other.m_value;
                }

                /// <summary>Comparison operator</summary>
                bool operator<(const MaxIterationCount &other) const
                {
                    return m_value < other.m_value;
                }

                /// <summary>Comparison operator</summary>
                bool operator>(const MaxIterationCount &other) const
                {
                    return m_value > other.m_value;
                }

                /// <summary>Comparison operator</summary>
                bool operator<=(const MaxIterationCount &other) const
                {
                    return m_value <= other.m_value;
                }

                /// <summary>Comparison operator</summary>
                bool operator>=(const MaxIterationCount &other) const
                {
                    return m_value >= other.m_value;
                }

                /// <summary>Operator to serialize the value to a stream</summary>
                friend std::ostream &operator<<(std::ostream &stream, const MaxIterationCount &value)
                {
                    return stream << value.toString();
                }

            private:
                void setFromString(const std::string &value);

                uint32_t m_value{ 200 };

                friend struct DataModel::Detail::Befriend<MaxIterationCount>;
            };

            using Descendants = std::tuple<
                LocalPointCloudRegistrationParameters::ConvergenceCriteria,
                LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold,
                LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold,
                LocalPointCloudRegistrationParameters::MaxCorrespondenceDistance,
                LocalPointCloudRegistrationParameters::MaxIterationCount>;

            /// <summary>Default constructor</summary>
            LocalPointCloudRegistrationParameters();

            /// <summary>Construct LocalPointCloudRegistrationParameters by loading from file</summary>
            explicit LocalPointCloudRegistrationParameters(const std::string &fileName);

            /// <summary>Construct a new LocalPointCloudRegistrationParameters instance from a previously serialized
            /// string</summary>
            /// <remarks>
            /// See also <see cref="serialize" />.
            /// </remarks>
            [[nodiscard]] static LocalPointCloudRegistrationParameters fromSerialized(const std::string &value);

            /// <summary>Serialize to a string</summary>
            /// <remarks>
            /// See also <see cref="fromSerialized" />. The returned string is formatted in YAML, and is compatible with
            /// the contents of files saved using <see cref="save" />.
            /// </remarks>
            std::string serialize() const;

            /// <summary>Constructor taking variadic number of arguments</summary>
            /// <remarks>
            /// One or more descendant types can be provided. All types not provided will be set to their default
            /// value. The result is the same as default constructing `LocalPointCloudRegistrationParameters` then invoking `set(args)`.
            ///
            /// The provided arguments must be descendants of `LocalPointCloudRegistrationParameters`. These types can be provided:
            ///
            /// <list type="bullet">
            /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria</item>
            /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold</item>
            /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold</item>
            /// <item>LocalPointCloudRegistrationParameters::MaxCorrespondenceDistance</item>
            /// <item>LocalPointCloudRegistrationParameters::MaxIterationCount</item>
            /// </list>
            /// </remarks>
#ifndef NO_DOC
            template<
                typename... Args,
                typename std::enable_if<sizeof...(Args) >= 1, int>::type = 0,
                typename std::enable_if<
                    Zivid::Detail::TypeTraits::AllArgsAreInTuple<Descendants, typename std::decay<Args>::type...>::
                        value,
                    int>::type = 0>
#else
            template<typename... Args>
#endif
            explicit LocalPointCloudRegistrationParameters(Args &&...args)
            {
                using namespace Zivid::Detail::TypeTraits;

                static_assert(
                    AllArgsDecayedAreUnique<Args...>::value,
                    "Found duplicate types among the arguments passed to LocalPointCloudRegistrationParameters(...). "
                    "Types should be listed at most once.");

                set(std::forward<Args>(args)...);
            }

            /// <summary>Set multiple arguments</summary>
            /// <remarks>
            /// The method invokes `set(arg)` with each of the provided arguments.
            ///
            /// The provided arguments must be descendants of `LocalPointCloudRegistrationParameters`. These types can be provided:
            ///
            /// <list type="bullet">
            /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria</item>
            /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold</item>
            /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold</item>
            /// <item>LocalPointCloudRegistrationParameters::MaxCorrespondenceDistance</item>
            /// <item>LocalPointCloudRegistrationParameters::MaxIterationCount</item>
            /// </list>
            /// </remarks>
#ifndef NO_DOC
            template<typename... Args, typename std::enable_if<sizeof...(Args) >= 2, int>::type = 0>
#else
            template<typename... Args>
#endif
            void set(Args &&...args)
            {
                using namespace Zivid::Detail::TypeTraits;

                using AllArgsAreDescendantNodes = AllArgsAreInTuple<Descendants, typename std::decay<Args>::type...>;
                static_assert(
                    AllArgsAreDescendantNodes::value, "All arguments passed to set(...) must be descendant nodes.");

                static_assert(
                    AllArgsDecayedAreUnique<Args...>::value,
                    "Found duplicate types among the arguments passed to set(...). "
                    "Types should be listed at most once.");

                Zivid::DataModel::Detail::invokeSetWithEachArgument(*this, std::forward<Args>(args)...);
            }

            /// <summary>Returns a copy of this object with the given argument(s) set to the new value(s)</summary>
            /// <remarks>
            /// Creates a copy of this LocalPointCloudRegistrationParameters object, then invokes `set(args)` on the copy, and
            /// finally returns the copy. This method does not modify the original object.
            ///
            /// The provided arguments must be descendants of `LocalPointCloudRegistrationParameters`. These types can be provided:
            ///
            /// <list type="bullet">
            /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria</item>
            /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold</item>
            /// <item>LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold</item>
            /// <item>LocalPointCloudRegistrationParameters::MaxCorrespondenceDistance</item>
            /// <item>LocalPointCloudRegistrationParameters::MaxIterationCount</item>
            /// </list>
            /// </remarks>
#ifndef NO_DOC
            template<typename... Args, typename std::enable_if<sizeof...(Args) >= 1, int>::type = 0>
#else
            template<typename... Args>
#endif
            LocalPointCloudRegistrationParameters copyWith(Args &&...args) const
            {
                using namespace Zivid::Detail::TypeTraits;

                using AllArgsAreDescendantNodes = AllArgsAreInTuple<Descendants, typename std::decay<Args>::type...>;
                static_assert(
                    AllArgsAreDescendantNodes::value,
                    "All arguments passed to copyWith(...) must be descendant nodes.");

                static_assert(
                    AllArgsDecayedAreUnique<Args...>::value,
                    "Found duplicate types among the arguments passed to copyWith(...). "
                    "Types should be listed at most once.");

                auto copy{ *this };
                copy.set(std::forward<Args>(args)...);
                return copy;
            }

            /// <summary>Get ConvergenceCriteria</summary>
            const ConvergenceCriteria &convergenceCriteria() const
            {
                return m_convergenceCriteria;
            }

            /// <summary>Get ConvergenceCriteria</summary>
            ConvergenceCriteria &convergenceCriteria()
            {
                return m_convergenceCriteria;
            }

            /// <summary>Set ConvergenceCriteria</summary>
            LocalPointCloudRegistrationParameters &set(const ConvergenceCriteria &value)
            {
                m_convergenceCriteria = value;
                return *this;
            }

            /// <summary>Set ConvergenceCriteria::RMSEDiffThreshold</summary>
            LocalPointCloudRegistrationParameters &set(const ConvergenceCriteria::RMSEDiffThreshold &value)
            {
                m_convergenceCriteria.set(value);
                return *this;
            }

            /// <summary>Set ConvergenceCriteria::SourceCoverageDiffThreshold</summary>
            LocalPointCloudRegistrationParameters &set(const ConvergenceCriteria::SourceCoverageDiffThreshold &value)
            {
                m_convergenceCriteria.set(value);
                return *this;
            }

            /// <summary>Get MaxCorrespondenceDistance</summary>
            const MaxCorrespondenceDistance &maxCorrespondenceDistance() const
            {
                return m_maxCorrespondenceDistance;
            }

            /// <summary>Get MaxCorrespondenceDistance</summary>
            MaxCorrespondenceDistance &maxCorrespondenceDistance()
            {
                return m_maxCorrespondenceDistance;
            }

            /// <summary>Set MaxCorrespondenceDistance</summary>
            LocalPointCloudRegistrationParameters &set(const MaxCorrespondenceDistance &value)
            {
                m_maxCorrespondenceDistance = value;
                return *this;
            }

            /// <summary>Get MaxIterationCount</summary>
            const MaxIterationCount &maxIterationCount() const
            {
                return m_maxIterationCount;
            }

            /// <summary>Get MaxIterationCount</summary>
            MaxIterationCount &maxIterationCount()
            {
                return m_maxIterationCount;
            }

            /// <summary>Set MaxIterationCount</summary>
            LocalPointCloudRegistrationParameters &set(const MaxIterationCount &value)
            {
                m_maxIterationCount = value;
                return *this;
            }

            template<
                typename T,
                typename std::enable_if<
                    std::is_same<T, LocalPointCloudRegistrationParameters::ConvergenceCriteria>::value,
                    int>::type = 0>
            const LocalPointCloudRegistrationParameters::ConvergenceCriteria &get() const
            {
                return m_convergenceCriteria;
            }

            template<
                typename T,
                typename std::enable_if<
                    std::is_same<T, LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold>::
                        value,
                    int>::type = 0>
            const LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold &get() const
            {
                return m_convergenceCriteria
                    .get<LocalPointCloudRegistrationParameters::ConvergenceCriteria::RMSEDiffThreshold>();
            }

            template<
                typename T,
                typename std::enable_if<
                    std::is_same<
                        T,
                        LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold>::value,
                    int>::type = 0>
            const LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold &get() const
            {
                return m_convergenceCriteria
                    .get<LocalPointCloudRegistrationParameters::ConvergenceCriteria::SourceCoverageDiffThreshold>();
            }

            template<
                typename T,
                typename std::enable_if<
                    std::is_same<T, LocalPointCloudRegistrationParameters::MaxCorrespondenceDistance>::value,
                    int>::type = 0>
            const LocalPointCloudRegistrationParameters::MaxCorrespondenceDistance &get() const
            {
                return m_maxCorrespondenceDistance;
            }

            template<
                typename T,
                typename std::enable_if<
                    std::is_same<T, LocalPointCloudRegistrationParameters::MaxIterationCount>::value,
                    int>::type = 0>
            const LocalPointCloudRegistrationParameters::MaxIterationCount &get() const
            {
                return m_maxIterationCount;
            }

            template<size_t i, typename std::enable_if<i == 0, int>::type = 0>
            const LocalPointCloudRegistrationParameters::ConvergenceCriteria &get() const
            {
                return m_convergenceCriteria;
            }

            template<size_t i, typename std::enable_if<i == 1, int>::type = 0>
            const LocalPointCloudRegistrationParameters::MaxCorrespondenceDistance &get() const
            {
                return m_maxCorrespondenceDistance;
            }

            template<size_t i, typename std::enable_if<i == 2, int>::type = 0>
            const LocalPointCloudRegistrationParameters::MaxIterationCount &get() const
            {
                return m_maxIterationCount;
            }

            /// <summary>Run the given function on each direct member with the value of the member as parameter</summary>
            template<typename F>
            void forEach(const F &f) const
            {
                f(m_convergenceCriteria);
                f(m_maxCorrespondenceDistance);
                f(m_maxIterationCount);
            }

            /// <summary>Run the given function on each direct member with the value of the member as parameter</summary>
            template<typename F>
            void forEach(const F &f)
            {
                f(m_convergenceCriteria);
                f(m_maxCorrespondenceDistance);
                f(m_maxIterationCount);
            }

            /// <summary>Equality operator</summary>
            bool operator==(const LocalPointCloudRegistrationParameters &other) const;

            /// <summary>Inequality operator</summary>
            bool operator!=(const LocalPointCloudRegistrationParameters &other) const;

            /// <summary>Get the value as string</summary>
            std::string toString() const;

            /// <summary>Operator to send the value as string to a stream</summary>
            friend std::ostream &operator<<(std::ostream &stream, const LocalPointCloudRegistrationParameters &value)
            {
                return stream << value.toString();
            }

            /// <summary>Save to the given file</summary>
            void save(const std::string &fileName) const;

            /// <summary>Load from the given file</summary>
            void load(const std::string &fileName);

        private:
            void setFromString(const std::string &value);

            void setFromString(const std::string &fullPath, const std::string &value);

            std::string getString(const std::string &fullPath) const;

            ConvergenceCriteria m_convergenceCriteria;
            MaxCorrespondenceDistance m_maxCorrespondenceDistance;
            MaxIterationCount m_maxIterationCount;

            friend struct DataModel::Detail::Befriend<LocalPointCloudRegistrationParameters>;
        };

#ifndef NO_DOC
        template<>
        struct LocalPointCloudRegistrationParameters::Version<1>
        {
            using Type = LocalPointCloudRegistrationParameters;
        };
#endif

    } // namespace Experimental
} // namespace Zivid

#ifndef NO_DOC
/// This namespace is for implementation details, and should not be relied on by users of the Zivid API.
namespace Zivid::Detail
{

    ZIVID_CORE_EXPORT void save(
        const Zivid::Experimental::LocalPointCloudRegistrationParameters &dataModel,
        std::ostream &ostream);
    ZIVID_CORE_EXPORT void load(
        Zivid::Experimental::LocalPointCloudRegistrationParameters &dataModel,
        std::istream &istream);

    ZIVID_CORE_EXPORT std::vector<uint8_t> serializeToBinaryVector(
        const Zivid::Experimental::LocalPointCloudRegistrationParameters &source);
    ZIVID_CORE_EXPORT void deserializeFromBinaryVector(
        Zivid::Experimental::LocalPointCloudRegistrationParameters &dest,
        const std::vector<uint8_t> &data);

} // namespace Zivid::Detail
#endif

#ifdef _MSC_VER
#    pragma warning(pop)
#endif

#ifndef NO_DOC
#    if !(defined(_MSC_VER) && (_MSC_VER <= 1900))
namespace std // NOLINT
{

    template<>
    struct tuple_size<Zivid::Experimental::LocalPointCloudRegistrationParameters::ConvergenceCriteria>
        : integral_constant<size_t, 2>
    {};

    template<size_t i>
    struct tuple_element<i, Zivid::Experimental::LocalPointCloudRegistrationParameters::ConvergenceCriteria>
    {
        static_assert(
            i < tuple_size<Zivid::Experimental::LocalPointCloudRegistrationParameters::ConvergenceCriteria>::value,
            "Index must be less than 2");

        using type // NOLINT
            = decltype(declval<Zivid::Experimental::LocalPointCloudRegistrationParameters::ConvergenceCriteria>()
                           .get<i>());
    };

    template<>
    struct tuple_size<Zivid::Experimental::LocalPointCloudRegistrationParameters> : integral_constant<size_t, 3>
    {};

    template<size_t i>
    struct tuple_element<i, Zivid::Experimental::LocalPointCloudRegistrationParameters>
    {
        static_assert(
            i < tuple_size<Zivid::Experimental::LocalPointCloudRegistrationParameters>::value,
            "Index must be less than 3");

        using type // NOLINT
            = decltype(declval<Zivid::Experimental::LocalPointCloudRegistrationParameters>().get<i>());
    };

} // namespace std
#    endif
#endif

// If we have access to the DataModel library, automatically include internal DataModel
// header. This header is necessary for serialization and deserialization.
#if defined(__has_include) && !defined(NO_DOC)
#    if __has_include(                                                                                                 \
        "Zivid/Experimental/LocalPointCloudRegistrationParametersInternal.h")                                          \
        && __has_include("Zivid/DataModelNodeMetaData.h")
#        include "Zivid/Experimental/LocalPointCloudRegistrationParametersInternal.h"
#    endif
#endif
