/*******************************************************************************
 * This file is part of the Zivid API
 *
 * Copyright 2015-2025 (C) Zivid AS
 * All rights reserved.
 *
 * Zivid Software License, v1.0
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of Zivid AS nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 4. This software, with or without modification, must not be used with any
 * other 3D camera than from Zivid AS.
 *
 * 5. Any software provided in binary form under this license must not be
 * reverse engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY ZIVID AS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL ZIVID AS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contact: Zivid Customer Success Team <customersuccess@zivid.com>
 * Info:    http://www.zivid.com
 ******************************************************************************/

#pragma once

#include <Zivid/Calibration/Pose.h>
#include <Zivid/Detail/CoreExport.h>
#include <Zivid/Detail/Pimpl.h>
#include <Zivid/Experimental/LocalPointCloudRegistrationParameters.h>
#include <Zivid/Matrix.h>
#include <Zivid/UnorganizedPointCloud.h>

#include <iosfwd>
#include <string>

namespace Zivid
{
    namespace Experimental
    {
        namespace Toolbox
        {
            using Pose = Zivid::Calibration::Pose;

            /// <summary>
            /// The result of a call to localPointCloudRegistration()
            /// </summary>
            class LocalPointCloudRegistrationResult
            {
                ZIVID_PIMPL_VALUE_SEMANTICS(LocalPointCloudRegistrationResult, ZIVID_CORE_EXPORT);

            public:
                /// <summary>The transform that must be applied to the source point cloud for it to align with the target point cloud</summary>
                ZIVID_CORE_EXPORT Pose transform() const;

                /// <summary>A boolean indicating whether the convergence criteria were satisfied before reaching the iteration limit</summary>
                ZIVID_CORE_EXPORT bool converged() const;

                /// <summary>The fraction of points in the source point cloud that has a correspondence in the target point cloud after transformation</summary>
                ZIVID_CORE_EXPORT float sourceCoverage() const;

                /// <summary>The root mean squared distance between corresponding points in the source and target point cloud after transformation</summary>
                ZIVID_CORE_EXPORT float rootMeanSquareError() const;

                /// <summary>Get string representation of the result</summary>
                ZIVID_CORE_EXPORT std::string toString() const;
            };

            /// <summary>Serialize the value to a stream</summary>
            ZIVID_CORE_EXPORT std::ostream &operator<<(
                std::ostream &stream,
                const LocalPointCloudRegistrationResult &result);

            /// <summary>Compute alignment transform between two point clouds</summary>
            /// <remarks>
            /// Given a `source` point cloud and a `target` point cloud, this function attempts to compute the transform
            /// that must be applied to the `source` in order to align it with the `target`. This can be used to create a
            /// "stitched" unorganized point cloud of an object by combining data collected from different camera angles.
            ///
            /// This function takes an argument `initialTransform` which is used as a starting-point for the computation
            /// of the transform that best aligns `source` with `target`. This initial guess is usually found from e.g.
            /// reference markers or robot capture pose, and this function is then used to refine the alignment. If the
            /// overlap of `source` and `target` is already quite good, one can pass the identity matrix as `initialTransform`.
            ///
            /// The returned transform represents the total transform needed to align `source` with `target`, i.e. it
            /// includes both `initialTransform` and the refinement found by the algorithm.
            ///
            /// Performance is very dependent on the number of points in either point cloud. To improve performance,
            /// voxel downsample one or both point clouds before passing them into this function. The resulting alignment
            /// transform can then be applied to the non-downsampled point clouds to still obtain a dense result.
            ///
            /// Performance is also very dependent on `MaxCorrespondenceDistance`. To improve performance, try reducing
            /// this value. However, keep the value larger than the typical point-to-point distance in the point clouds,
            /// and larger than the expected translation error in the initial guess.
            /// </remarks>
            /// <param name="target">The point cloud to align with</param>
            /// <param name="source">The point cloud to be aligned with target</param>
            /// <param name="params">Parameters for the registration process and its convergence criteria</param>
            /// <param name="initialTransform">Initial guess applied to source point cloud before refinement</param>
            /// <returns>Instance of LocalPointCloudRegistrationResult</returns>
            ZIVID_CORE_EXPORT
            LocalPointCloudRegistrationResult localPointCloudRegistration(
                const UnorganizedPointCloud &target,
                const UnorganizedPointCloud &source,
                const LocalPointCloudRegistrationParameters &params,
                const Pose &initialTransform = Pose{ Matrix4x4::identity() });
        } // namespace Toolbox
    } // namespace Experimental
} // namespace Zivid
