/*******************************************************************************
 * This file is part of the Zivid API
 *
 * Copyright 2015-2025 (C) Zivid AS
 * All rights reserved.
 *
 * Zivid Software License, v1.0
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of Zivid AS nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 4. This software, with or without modification, must not be used with any
 * other 3D camera than from Zivid AS.
 *
 * 5. Any software provided in binary form under this license must not be
 * reverse engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY ZIVID AS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL ZIVID AS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contact: Zivid Customer Success Team <customersuccess@zivid.com>
 * Info:    http://www.zivid.com
 ******************************************************************************/

#pragma once

#include "Zivid/Array1D.h"
#include "Zivid/Color.h"
#include "Zivid/Detail/CoreExport.h"
#include "Zivid/Matrix.h"
#include "Zivid/Point.h"
#include "Zivid/SNR.h"

#include <iosfwd>
#include <memory>
#include <optional>
#include <string>

namespace Zivid
{
    /// <summary>
    /// Point cloud with x, y, z, RGB color and SNR laid out as a linear list of only valid points
    /// </summary>
    /// <remarks>
    /// An instance of this class is a handle to a point cloud stored on the compute device memory.
    /// This class provides several methods to copy point cloud data from the compute device
    /// memory to host (CPU) system memory (RAM).
    ///
    /// This point cloud contains only valid points, meaning that the XYZ values are never NaN.
    ///
    /// </remarks>
    class UnorganizedPointCloud
    {
    public:
        /// <summary>Create an empty point cloud</summary>
        /// <remarks>
        /// This constructor creates a point cloud with size zero, i.e. no points. An empty point cloud
        /// can be useful for combining points from several other points clouds.
        /// </remarks>
        ZIVID_CORE_EXPORT UnorganizedPointCloud();

        /// <summary>Get the size of the point cloud (total number of points)</summary>
        ZIVID_CORE_EXPORT size_t size() const;

        /// <summary>Create a new point cloud containing the combined data of this point cloud and another</summary>
        /// <param name="other">The other point cloud to copy data from</param>
        /// <returns>A new point cloud containing the combined data</returns>
        ZIVID_CORE_EXPORT UnorganizedPointCloud extended(const UnorganizedPointCloud &other) const;

        /// <summary>Extend this point cloud in-place by adding the points from another point cloud</summary>
        /// <param name="other">The other point cloud to copy data from</param>
        ZIVID_CORE_EXPORT UnorganizedPointCloud &extend(const UnorganizedPointCloud &other);

        /// <summary>Create a new point cloud that is a voxel downsampling of this point cloud</summary>
        /// <remarks>
        /// Voxel downsampling subdivides 3D space into a grid of cubic voxels with a given size. If a given voxel
        /// contains a number of points at or above the given limit, all those source points are replaced with a single
        /// point with the following properties:
        /// <list type="bullet">
        ///    <item>
        ///        Position (XYZ) is an SNR-weighted average of the source points' positions, i.e. a high-confidence
        ///        source point will have a greater influence on the resulting position than a low-confidence one.
        ///    </item>
        ///    <item>
        ///        Color (RGBA) is the average of the source points' colors.
        ///    </item>
        ///    <item>
        ///        Signal-to-noise ratio (SNR) is sqrt(sum(SNR^2)) of the source points' SNR values, i.e. the SNR of
        ///        a new point will increase with both the number and the confidence of the source points that were used
        ///        to compute its position.
        ///    </item>
        /// </list>
        ///
        /// Using `minPointsPerVoxel > 1` is particularly useful for removing noise and artifacts from unorganized
        /// point clouds that are a combination of structured point clouds captured from different angles. This is
        /// because a given artifact is most likely only present in one of the captures, and `minPointsPerVoxel` can be
        /// used to only fill voxels that both captures "agree" on.
        /// </remarks>
        /// <param name="voxelSize">The size of the voxel cubes (must be greater than 0.0)</param>
        /// <param name="minPointsPerVoxel">The minimum number of points required to fill a voxel (must be 1 or greater)</param>
        /// <returns>A new point cloud containing the voxel downsampled data</returns>
        ZIVID_CORE_EXPORT UnorganizedPointCloud voxelDownsampled(float voxelSize, int minPointsPerVoxel) const;

        /// <summary>Transform the point cloud in-place by the given 4x4 transformation matrix</summary>
        /// <remarks>The transform matrix must be affine. In other words, the last row of the matrix
        /// must be `[0, 0, 0, 1]`.</remarks>
        /// <param name="matrix">An affine 4x4 matrix</param>
        ZIVID_CORE_EXPORT UnorganizedPointCloud &transform(const Zivid::Matrix4x4 &matrix);

        /// <summary>Transform the point cloud by the given 4x4 transformation matrix</summary>
        /// <remarks>
        /// This method is identical to <see cref="transform" />, except the transformed point cloud is
        /// returned as a new UnorganizedPointCloud instance. The current point cloud is not modified.
        /// </remarks>
        /// <param name="matrix">An affine 4x4 matrix</param>
        ZIVID_CORE_EXPORT UnorganizedPointCloud transformed(const Zivid::Matrix4x4 &matrix) const;

        /// <summary>Translate the point cloud in-place so that its centroid lands at the origin (0,0,0)</summary>
        /// <remarks>If the point cloud has zero points, this method throws an exception.</remarks>
        ZIVID_CORE_EXPORT UnorganizedPointCloud &center();

        /// <summary>Get the centroid of the point cloud, i.e. average of all XYZ point positions</summary>
        /// <remarks>If the point cloud has zero points, this method returns an empty std::optional.</remarks>
        /// <returns>The centroid as an optional <see cref="PointXYZ" /></returns>
        ZIVID_CORE_EXPORT std::optional<PointXYZ> centroid() const;

        /// <summary>Get the XYZ data of every point</summary>
        /// <remarks>
        /// This method copies the 3D point coordinates from compute device memory to host
        /// (CPU) system memory (RAM) and returns them as an <see cref="Array1D" /> of <see cref="PointXYZ" />.
        ///
        /// The length of the returned array will match the \ref size method.
        ///
        /// If you already have a buffer, you can copy data directly from compute device
        /// memory to your destination buffer using \ref copyData with template parameter
        /// <see cref="PointXYZ" />.
        /// </remarks>
        ZIVID_CORE_EXPORT Array1D<PointXYZ> copyPointsXYZ() const;

        /// <summary>Get the color data of every point on 8-bit RGBA format</summary>
        /// <remarks>
        /// This method copies the RGBA values from compute device memory to host (CPU) system
        /// memory (RAM) and returns them as an <see cref="Array1D" /> of <see cref="ColorRGBA" />.
        ///
        /// The length of the returned array will match the \ref size method.
        ///
        /// If you already have a buffer, you can copy data directly from compute device
        /// memory to your destination buffer using \ref copyData with template parameter
        /// <see cref="ColorRGBA" />.
        /// </remarks>
        ZIVID_CORE_EXPORT Array1D<ColorRGBA> copyColorsRGBA() const;

        /// <summary>Get the color data of every point on 8-bit BGRA format</summary>
        /// <remarks>
        /// This method copies the BGRA values from compute device memory to host (CPU) system
        /// memory (RAM) and returns them as an <see cref="Array1D" /> of <see cref="ColorBGRA" />.
        ///
        /// The length of the returned array will match the \ref size method.
        ///
        /// If you already have a buffer, you can copy data directly from compute device
        /// memory to your destination buffer using \ref copyData with template parameter
        /// <see cref="ColorBGRA" />.
        /// </remarks>
        ZIVID_CORE_EXPORT Array1D<ColorBGRA> copyColorsBGRA() const;

        /// <summary>Get the color data of every point on 8-bit RGBA format in the sRGB color space</summary>
        /// <remarks>
        /// This method copies the sRGB values from compute device memory to host (CPU) system
        /// memory (RAM) and returns them as an <see cref="Array1D" /> of <see cref="ColorRGBA_SRGB" />.
        ///
        /// The sRGB color space is suitable for showing an image on a display for human viewing. It is easier to
        /// see details in darker areas of an image in sRGB than in linear RGB, as more of the dynamic range
        /// is dedicated to darker colors. This color space is assumed by default by most monitors and should be used when
        /// displaying an image.
        /// </remarks>
        ZIVID_CORE_EXPORT Array1D<ColorRGBA_SRGB> copyColorsRGBA_SRGB() const;

        /// <summary>Get the color data of every point on 8-bit BGRA format in the sRGB color space</summary>
        /// <remarks>
        /// This method copies the sRGB values from compute device memory to host (CPU) system
        /// memory (RAM) and returns them as an <see cref="Array1D" /> of <see cref="ColorBGRA_SRGB" />.
        ///
        /// The sRGB color space is suitable for showing an image on a display for human viewing. It is easier to
        /// see details in darker areas of an image in sRGB than in linear RGB, as more of the dynamic range
        /// is dedicated to darker colors. This color space is assumed by default by most monitors and should be used when
        /// displaying an image.
        /// </remarks>
        ZIVID_CORE_EXPORT Array1D<ColorBGRA_SRGB> copyColorsBGRA_SRGB() const;

        /// <summary>Get the SNR data of every point</summary>
        /// <remarks>
        /// This method copies the signal-to-noise ratios from compute device memory to host (CPU) system
        /// memory (RAM) and returns them as an <see cref="Array1D" /> of <see cref="SNR" />.
        ///
        /// The length of the returned array will match the \ref size method.
        ///
        /// If you already have a buffer, you can copy data directly from compute device
        /// memory to your destination buffer using \ref copyData with template parameter
        /// <see cref="SNR" />.
        /// </remarks>
        ZIVID_CORE_EXPORT Array1D<SNR> copySNRs() const;

        /// <summary>Array1D with point cloud data using specified DataFormat</summary>
        /// <remarks>
        /// The template parameter DataFormat specifies which point cloud data fields are included
        /// in the array. This is a convenience function that can be used for generic template
        /// programming. Calling this method with template parameter <see cref="PointXYZ" />, for
        /// example, is equivalent to calling the <see cref="copyPointsXYZ" /> member function.
        ///
        /// If you already have a buffer, you can copy data directly from compute device memory to
        /// your destination buffer using `copyData(DataFormat *destination)`. For more information
        /// about the available DataFormat types, see `copyData(DataFormat *destination)`.
        /// </remarks>
        template<typename DataFormat>
        Array1D<DataFormat> copyData() const
        {
            static_assert(
                HasCopyDataImplReturnArray1DTag<DataFormat>::value,
                "The provided DataFormat is unsupported. Please refer to the documentation for "
                "this method for the supported DataFormats.");
            return copyDataImpl(ReturnArray1DTag<DataFormat>{});
        }

        /// <summary>Copy data in the specified DataFormat to a destination buffer</summary>
        /// <remarks>
        /// Transfers point cloud data from the compute device to a provided output buffer on the host
        /// (CPU) system memory (RAM). The template parameter DataFormat defines the layout of the data
        /// and which data fields are included.
        ///
        /// This method is the most efficient way to copy point cloud data from the compute device
        /// if you need to store it to your own buffer. The destination buffer must be large enough to hold
        /// <see cref="UnorganizedPointCloud::size()" /> * `sizeof(DataFormat)` bytes of data. Providing
        /// a buffer which is too small results in undefined behavior.
        ///
        /// Available data formats:
        /// <list type="bullet">
        ///    <item>
        ///        <see cref="PointXYZ" />: 3D coordinates. Contains 3 floats (`x`, `y`, `z`) packed
        ///        together. Size per element is 12 bytes.
        ///    </item>
        ///    <item>
        ///        <see cref="ColorRGBA" />: RGBA colors. Contains 4 uint8_t `r`, `g`, `b` and `a`.
        ///        `a` is always 255. Size per element 4 bytes.
        ///    </item>
        ///    <item>
        ///        <see cref="ColorBGRA" />: BGRA colors. Contains 4 uint8_t `b`, `g`, `r` and `a`.
        ///        `a` is always 255. Size per element 4 bytes.
        ///    </item>
        ///    <item>
        ///        <see cref="ColorRGBA_SRGB" />: RGBA colors in the sRGB color space. Contains 4 uint8_t `r`, `g`, `b`
        ///        and `a`. `a` is always 255. Size per element 4 bytes.
        ///    </item>
        ///    <item>
        ///        <see cref="ColorBGRA_SRGB" />: BGRA colors in the sRGB color space. Contains 4 uint8_t `b`, `g`, `r`
        ///        and `a`. `a` is always 255. Size per element 4 bytes.
        ///    </item>
        ///    <item>
        ///        <see cref="SNR" />: SNR values, contains one float member `value`.
        ///        Size per element is 4 bytes.
        ///    </item>
        /// </list>
        /// </remarks>
        /// <seealso cref="copyPointsXYZ" />
        /// <seealso cref="copyColorsRGBA" />
        /// <seealso cref="copyColorsBGRA" />
        /// <seealso cref="copyColorsRGBA_SRGB" />
        /// <seealso cref="copyColorsBGRA_SRGB" />
        /// <seealso cref="copySNRs" />
        template<typename DataFormat>
        void copyData(DataFormat *destination) const
        {
            static_assert(
                HasCopyDataImpl<DataFormat *>::value,
                "The provided DataFormat is unsupported. Please refer to the documentation for "
                "this method for the supported data formats.");
            return copyDataImpl(destination);
        }

        /// <summary>
        /// Returns a clone of the point cloud. The clone will include a copy of all of the point cloud
        /// data on the compute device memory. This means that the returned point cloud will not be
        /// affected by subsequent modifications (such as transform) on the original point cloud.
        /// </summary>
        /// <remarks>
        /// This function incurs a performance cost due to the copying of the compute device memory. When
        /// performance is important we recommend to avoid using this method, and instead modify the
        /// existing point cloud.
        /// </remarks>
        ZIVID_CORE_EXPORT UnorganizedPointCloud clone() const;

        /// <summary>Get string representation of the point cloud</summary>
        /// <returns>Point cloud info as string</returns>
        ZIVID_CORE_EXPORT std::string toString() const;

#ifndef NO_DOC
        ZIVID_CORE_EXPORT class UnorganizedPointCloudImpl &getImpl();
        ZIVID_CORE_EXPORT const class UnorganizedPointCloudImpl &getImpl() const;
        ZIVID_CORE_EXPORT explicit UnorganizedPointCloud(std::shared_ptr<class UnorganizedPointCloudImpl> other);
#endif

    private:
        template<class T>
        struct ReturnArray1DTag
        {};

        template<typename... Args>
        class HasCopyDataImplReturnArray1DTag
        {
            template<
                typename C,
                typename = decltype(std::declval<C>().copyDataImpl(std::declval<ReturnArray1DTag<Args>>()...))>
            static std::true_type test(int);
            template<typename C>
            static std::false_type test(...);

        public:
            static constexpr bool value = decltype(test<UnorganizedPointCloud>(0))::value;
        };

        template<typename... Args>
        class HasCopyDataImpl
        {
            template<typename C, typename = decltype(std::declval<C>().copyDataImpl(std::declval<Args>()...))>
            static std::true_type test(int);
            template<typename C>
            static std::false_type test(...);

        public:
            static constexpr bool value = decltype(test<UnorganizedPointCloud>(0))::value;
        };

        ZIVID_CORE_EXPORT Array1D<PointXYZ> copyDataImpl(ReturnArray1DTag<PointXYZ> /*tag*/) const;
        ZIVID_CORE_EXPORT Array1D<ColorRGBA> copyDataImpl(ReturnArray1DTag<ColorRGBA> /*tag*/) const;
        ZIVID_CORE_EXPORT Array1D<ColorBGRA> copyDataImpl(ReturnArray1DTag<ColorBGRA> /*tag*/) const;
        ZIVID_CORE_EXPORT Array1D<ColorRGBA_SRGB> copyDataImpl(ReturnArray1DTag<ColorRGBA_SRGB> /*tag*/) const;
        ZIVID_CORE_EXPORT Array1D<ColorBGRA_SRGB> copyDataImpl(ReturnArray1DTag<ColorBGRA_SRGB> /*tag*/) const;
        ZIVID_CORE_EXPORT Array1D<SNR> copyDataImpl(ReturnArray1DTag<SNR> /*tag*/) const;

        ZIVID_CORE_EXPORT void copyDataImpl(PointXYZ *destination) const;
        ZIVID_CORE_EXPORT void copyDataImpl(ColorRGBA *destination) const;
        ZIVID_CORE_EXPORT void copyDataImpl(ColorBGRA *destination) const;
        ZIVID_CORE_EXPORT void copyDataImpl(ColorRGBA_SRGB *destination) const;
        ZIVID_CORE_EXPORT void copyDataImpl(ColorBGRA_SRGB *destination) const;
        ZIVID_CORE_EXPORT void copyDataImpl(SNR *destination) const;

        std::shared_ptr<class UnorganizedPointCloudImpl> m_impl;
    };

    /// <summary>Serialize the value to a stream</summary>
    ZIVID_CORE_EXPORT std::ostream &operator<<(
        std::ostream &stream,
        const UnorganizedPointCloud &unorganizedPointCloud);

#ifndef NO_DOC
    /// This namespace is for implementation details, and should not be relied on by users of the Zivid API.
    namespace Detail
    {
        ZIVID_CORE_EXPORT void waitUntilProcessingIsComplete(const UnorganizedPointCloud &unorganizedPointCloud);
    } // namespace Detail
#endif
} // namespace Zivid
